@@ -19,6 +19,7 @@ from typing import (
     TYPE_CHECKING,
     Any,
     Deque,
+    Iterator,
     Literal,
     NamedTuple,
     Optional,
@@ -26,13 +27,14 @@ from typing import (
 )
 
 from .child import ProcessDesc
-from .cli_stub import CLIOptions
+from .cli_stub import CLIOptions, SaveAsSessionOptions
 from .clipboard import ClipboardRequestManager, set_clipboard_string
 from .constants import (
     appname,
     clear_handled_signals,
     config_dir,
     kitten_exe,
+    unserialize_launch_flag,
     wakeup_io_loop,
 )
 from .fast_data_types import (
@@ -67,6 +69,7 @@ from .fast_data_types import (
     get_click_interval,
     get_mouse_data_for_window,
     get_options,
+    get_window_logo_settings_if_not_default,
     is_css_pointer_name_valid,
     is_modifier_key,
     last_focused_os_window_id,
@@ -88,11 +91,13 @@ from .fast_data_types import (
     wakeup_main_loop,
 )
 from .keys import keyboard_mode_name, mod_mask
+from .notifications import NotificationManager
+from .options.types import Options
 from .progress import Progress
 from .rgb import to_color
 from .terminfo import get_capabilities
 from .types import MouseEvent, OverlayType, WindowGeometry, ac, run_once
-from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
+from .typing_compat import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
     color_as_int,
     docs_url,
@@ -107,7 +112,7 @@ from .utils import (
     sanitize_control_codes,
     sanitize_for_bracketed_paste,
     sanitize_title,
-    sanitize_url_for_dispay_to_user,
+    sanitize_url_for_display_to_user,
     shlex_split,
 )
 
@@ -154,7 +159,7 @@ class CwdRequest:
             return window.get_cwd_of_root_child() or ''
         return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
 
-    def modify_argv_for_launch_with_cwd(self, argv: list[str], env: dict[str, str] | None=None) -> str:
+    def modify_argv_for_launch_with_cwd(self, argv: list[str], env: dict[str, str] | None=None, hold_after_ssh: bool = False) -> str:
         window = self.window
         if not window:
             return ''
@@ -165,11 +170,13 @@ class CwdRequest:
                 run_shell = argv[0] == resolved_shell(get_options())[0]
                 server_args = [] if run_shell else list(argv)
                 from kittens.ssh.utils import set_cwd_in_cmdline, set_env_in_cmdline, set_server_args_in_cmdline
+                if ssh_kitten_cmdline and ssh_kitten_cmdline[0] == 'kitten':
+                    ssh_kitten_cmdline[0] = kitten_exe()
                 argv[:] = ssh_kitten_cmdline
-                if argv and argv[0] == 'kitten':
-                    argv[0] = kitten_exe()
                 set_cwd_in_cmdline(reported_cwd, argv)
                 set_server_args_in_cmdline(server_args, argv, allocate_tty=not run_shell)
+                if hold_after_ssh:
+                    argv[:0] = [kitten_exe(), "run-shell"]
                 if env is not None:
                     # Assume env is coming from a local process so drop env
                     # vars that can cause issues when set on the remote host
@@ -242,6 +249,7 @@ class WindowDict(TypedDict):
     user_vars: dict[str, str]
     at_prompt: bool
     created_at: int
+    in_alternate_screen: bool
 
 
 class PipeData(TypedDict):
@@ -293,6 +301,7 @@ class Watchers:
     on_title_change: list[Watcher]
     on_cmd_startstop: list[Watcher]
     on_color_scheme_preference_change: list[Watcher]
+    on_tab_bar_dirty: list[Watcher]
 
     def __init__(self) -> None:
         self.on_resize = []
@@ -302,6 +311,7 @@ class Watchers:
         self.on_title_change = []
         self.on_cmd_startstop = []
         self.on_color_scheme_preference_change = []
+        self.on_tab_bar_dirty = []
 
     def add(self, others: 'Watchers') -> None:
         def merge(base: list[Watcher], other: list[Watcher]) -> None:
@@ -315,11 +325,13 @@ class Watchers:
         merge(self.on_title_change, others.on_title_change)
         merge(self.on_cmd_startstop, others.on_cmd_startstop)
         merge(self.on_color_scheme_preference_change, others.on_color_scheme_preference_change)
+        merge(self.on_tab_bar_dirty, others.on_tab_bar_dirty)
 
     def clear(self) -> None:
         del self.on_close[:], self.on_resize[:], self.on_focus_change[:]
         del self.on_set_user_var[:], self.on_title_change[:], self.on_cmd_startstop[:]
         del self.on_color_scheme_preference_change[:]
+        del self.on_tab_bar_dirty[:]
 
     def copy(self) -> 'Watchers':
         ans = Watchers()
@@ -330,12 +342,13 @@ class Watchers:
         ans.on_title_change = self.on_title_change[:]
         ans.on_cmd_startstop = self.on_cmd_startstop[:]
         ans.on_color_scheme_preference_change = self.on_color_scheme_preference_change[:]
+        ans.on_tab_bar_dirty = self.on_tab_bar_dirty[:]
         return ans
 
     @property
     def has_watchers(self) -> bool:
         return bool(self.on_close or self.on_resize or self.on_focus_change or self.on_color_scheme_preference_change
-                    or self.on_set_user_var or self.on_title_change or self.on_cmd_startstop)
+                    or self.on_set_user_var or self.on_title_change or self.on_cmd_startstop or self.on_tab_bar_dirty)
 
 
 def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data: dict[str, Any]) -> None:
@@ -349,6 +362,31 @@ def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data:
     add_timer(callback, 0, False)
 
 
+class WindowCreationSpec(NamedTuple):
+    use_shell: bool = True
+    cmd: list[str] | None = None
+    has_stdin: bool = False
+    override_title: str | None = None
+    cwd_from: CwdRequest | None = None
+    cwd: str | None = None
+    overlay_for: int | None = None
+    env: tuple[tuple[str, str], ...] | None = None
+    location: str | None = None
+    copy_colors_from: int | None = None
+    colors: tuple[str, ...] = ()
+    allow_remote_control: bool = False
+    marker: str | None = None
+    watchers: tuple[str, ...] = ()
+    overlay_behind: bool = False
+    is_clone_launch: str = ''
+    remote_control_passwords: dict[str, Sequence[str]] | None = None
+    hold: bool = False
+    bias: float | None = None
+    hold_after_ssh: bool = False
+    spacing: tuple[str, ...] = ()
+    user_vars: tuple[tuple[str, str], ...] = ()
+
+
 def pagerhist(screen: Screen, as_ansi: bool = False, add_wrap_markers: bool = True, upto_output_start: bool = False) -> str:
     pht = screen.historybuf.pagerhist_as_text(upto_output_start)
     if pht and (not as_ansi or not add_wrap_markers):
@@ -543,6 +581,14 @@ def color_control(cp: ColorProfile, code: int, value: str | bytes | memoryview =
     return ''
 
 
+def da1(opts: Options) -> str:
+    ans = '?62;'
+    if 'write-clipboard' in opts.clipboard_control:
+        # see https://github.com/contour-terminal/vt-extensions/blob/master/clipboard-extension.md
+        ans += '52;'
+    return ans + 'c'
+
+
 class EdgeWidths:
     left: float | None
     top: float | None
@@ -564,6 +610,16 @@ class EdgeWidths:
     def copy(self) -> 'EdgeWidths':
         return EdgeWidths(self.serialize())
 
+    def as_launch_args(self, prefix: str = 'padding') -> Iterator[str]:
+        if self.left is not None:
+            yield f'--spacing={prefix}-left={self.left}'
+        if self.right is not None:
+            yield f'--spacing={prefix}-left={self.right}'
+        if self.top is not None:
+            yield f'--spacing={prefix}-left={self.top}'
+        if self.bottom is not None:
+            yield f'--spacing={prefix}-left={self.bottom}'
+
 
 class GlobalWatchers:
 
@@ -597,6 +653,9 @@ class Window:
     overlay_type = OverlayType.transient
     initial_ignore_focus_changes: bool = False
     initial_ignore_focus_changes_context_manager_in_operation: bool = False
+    creation_spec: WindowCreationSpec | None = None
+    created_in_session_name: str = ''
+    serialized_id: int = 0
 
     @classmethod
     @contextmanager
@@ -632,10 +691,10 @@ class Window:
         self.last_focused_at = 0.
         self.is_focused: bool = False
         self.progress = Progress()
+        self.clear_progress_timer: int = 0
         self.last_resized_at = 0.
         self.started_at = monotonic()
         self.created_at = time_ns()
-        self.clear_progress_timer: int = 0
         self.current_remote_data: list[str] = []
         self.current_mouse_event_button = 0
         self.current_clipboard_read_ask: bool | None = None
@@ -659,12 +718,12 @@ class Window:
         self.title_stack: Deque[str] = deque(maxlen=10)
         self.user_vars: dict[str, str] = {}
         self.id: int = add_window(tab.os_window_id, tab.id, self.title)
+        if not self.id:
+            raise Exception(f'No tab with id: {tab.id} in OS Window: {tab.os_window_id} was found, or the window counter wrapped')
         self.clipboard_request_manager = ClipboardRequestManager(self.id)
         self.margin = EdgeWidths()
         self.padding = EdgeWidths()
         self.kitten_result: dict[str, Any] | None = None
-        if not self.id:
-            raise Exception(f'No tab with id: {tab.id} in OS Window: {tab.os_window_id} was found, or the window counter wrapped')
         self.tab_id = tab.id
         self.os_window_id = tab.os_window_id
         self.tabref: Callable[[], TabType | None] = weakref.ref(tab)
@@ -762,54 +821,6 @@ class Window:
     def __repr__(self) -> str:
         return f'Window(title={self.title}, id={self.id})'
 
-    def as_dict(self, is_focused: bool = False, is_self: bool = False, is_active: bool = False) -> WindowDict:
-        return {
-            'id': self.id,
-            'is_focused': is_focused,
-            'is_active': is_active,
-            'title': self.title,
-            'pid': self.child.pid,
-            'cwd': self.child.current_cwd or self.child.cwd,
-            'cmdline': self.child.cmdline,
-            'last_reported_cmdline': self.last_cmd_cmdline,
-            'last_cmd_exit_status': self.last_cmd_exit_status,
-            'env': self.child.environ or self.child.final_env,
-            'foreground_processes': self.child.foreground_processes,
-            'is_self': is_self,
-            'at_prompt': self.at_prompt,
-            'lines': self.screen.lines,
-            'columns': self.screen.columns,
-            'user_vars': self.user_vars,
-            'created_at': self.created_at,
-        }
-
-    def serialize_state(self) -> dict[str, Any]:
-        ans = {
-            'version': 1,
-            'id': self.id,
-            'child_title': self.child_title,
-            'override_title': self.override_title,
-            'default_title': self.default_title,
-            'title_stack': list(self.title_stack),
-            'allow_remote_control': self.allow_remote_control,
-            'remote_control_passwords': self.remote_control_passwords,
-            'cwd': self.child.current_cwd or self.child.cwd,
-            'env': self.child.environ,
-            'cmdline': self.child.cmdline,
-            'last_reported_cmdline': self.last_cmd_cmdline,
-            'last_cmd_exit_status': self.last_cmd_exit_status,
-            'margin': self.margin.serialize(),
-            'user_vars': self.user_vars,
-            'padding': self.padding.serialize(),
-        }
-        if self.window_custom_type:
-            ans['window_custom_type'] = self.window_custom_type
-        if self.overlay_type is not OverlayType.transient:
-            ans['overlay_type'] = self.overlay_type.value
-        if self.user_vars:
-            ans['user_vars'] = self.user_vars
-        return ans
-
     @property
     def overlay_parent(self) -> Optional['Window']:
         tab = self.tabref()
@@ -829,7 +840,7 @@ class Window:
     def has_running_program(self) -> bool:
         return not self.at_prompt
 
-    def matches(self, field: str, pat: MatchPatternType) -> bool:
+    def matches(self, field: str, pat: MatchPatternType, active_session: str, most_recent_session: str) -> bool:
         if isinstance(pat, tuple):
             if field == 'env':
                 return key_val_matcher(self.child.environ.items(), *pat)
@@ -850,9 +861,21 @@ class Window:
                 if pat.search(x) is not None:
                     return True
             return False
+        if field == 'session':
+            match pat.pattern:
+                case '.':
+                    return self.created_in_session_name == active_session
+                case '~':
+                    return self.created_in_session_name == active_session or self.created_in_session_name == most_recent_session
+
+            return pat.search(self.created_in_session_name) is not None
         return False
 
-    def matches_query(self, field: str, query: str, active_tab: TabType | None = None, self_window: Optional['Window'] = None) -> bool:
+    def matches_query(
+        self, field: str, query: str, active_tab: TabType | None = None,
+        self_window: Optional['Window'] = None, active_session: str = '',
+        most_recent_session: str = '',
+    ) -> bool:
         if field in ('num', 'recent'):
             if active_tab is not None:
                 try:
@@ -880,6 +903,8 @@ class Window:
                 return False
             if query == 'parent_focused':
                 return active_tab is not None and self.tabref() is active_tab and last_focused_os_window_id() == self.os_window_id
+            if query == 'focused_os_window':
+                return last_focused_os_window_id() == self.os_window_id
             if query == 'self':
                 return self is self_window
             if query == 'overlay_parent':
@@ -901,7 +926,7 @@ class Window:
             return gid is not None and t.windows.active_window_in_group_id(gid) is self
 
         pat = compile_match_query(query, field not in ('env', 'var'))
-        return self.matches(field, pat)
+        return self.matches(field, pat, active_session, most_recent_session)
 
     def set_visible_in_layout(self, val: bool) -> None:
         val = bool(val)
@@ -948,7 +973,9 @@ class Window:
             mark_os_window_dirty(self.os_window_id)
 
         self.geometry = g = new_geometry
-        set_window_render_data(self.os_window_id, self.tab_id, self.id, self.screen, *g[:4])
+        set_window_render_data(self.os_window_id, self.tab_id, self.id, self.screen,
+                             g.left, g.top, g.right, g.bottom,
+                             g.spaces.left, g.spaces.top, g.spaces.right, g.spaces.bottom)
         self.update_effective_padding()
         if update_ime_position:
             update_ime_position_for_window(self.id, True)
@@ -1030,7 +1057,7 @@ class Window:
         text = ''.join(strings)
         get_boss().display_scrollback(self, text, title='Dump of lines', report_cursor=False)
 
-    def write_to_child(self, data: str | bytes) -> None:
+    def write_to_child(self, data: str | bytes | memoryview) -> None:
         if data:
             if isinstance(data, str):
                 data = data.encode('utf-8')
@@ -1178,7 +1205,7 @@ class Window:
             if opts.allow_hyperlinks & 0b10:
                 from kittens.tui.operations import styled
                 boss.choose(
-                    'What would you like to do with this URL:\n' + styled(sanitize_url_for_dispay_to_user(url), fg='yellow'),
+                    'What would you like to do with this URL:\n' + styled(sanitize_url_for_display_to_user(url), fg='yellow'),
                     partial(self.hyperlink_open_confirmed, url, cwd),
                     'o:Open', 'c:Copy to clipboard', 'n;red:Nothing', default='o',
                     window=self, title=_('Hyperlink activated'),
@@ -1277,6 +1304,9 @@ class Window:
             return True
         return False
 
+    def on_da1(self) -> None:
+        self.screen.send_escape_code_to_child(ESC_CSI, da1(get_options()))
+
     def on_bell(self) -> None:
         cb = get_options().command_on_bell
         if cb and cb != ['none']:
@@ -1305,6 +1335,9 @@ class Window:
         b |= b << 8
         self.screen.send_escape_code_to_child(ESC_OSC, f'{code};rgb:{r:04x}/{g:04x}/{b:04x}')
 
+    def on_reset(self) -> None:
+        pass
+
     def notify_child_of_resize(self) -> None:
         pty_size = self.last_reported_pty_size
         if pty_size[0] > -1 and self.screen.in_band_resize_notification:
@@ -1447,8 +1480,7 @@ class Window:
         self.kitten_result_processors.append(callback)
 
     def handle_overlay_ready(self, msg: memoryview) -> None:
-        boss = get_boss()
-        tab = boss.tab_for_window(self)
+        tab = self.tabref()
         if tab is not None:
             tab.move_window_to_top_of_group(self)
         if self.keys_redirected_till_ready_from:
@@ -1585,16 +1617,23 @@ class Window:
             cmd.only_when = OnlyWhen(when)
             if not nm.is_notification_allowed(cmd, self.id):
                 return
-            if action == 'notify':
-                if self.last_cmd_end_notification is not None:
+
+            def notify(window: Window, opts: Options, nm: NotificationManager) -> None:
+                if window.last_cmd_end_notification is not None:
                     if 'next' in opts.notify_on_cmd_finish.clear_on:
-                        nm.close_notification(self.last_cmd_end_notification[0])
-                    self.last_cmd_end_notification = None
-                notification_id = nm.notify_with_command(cmd, self.id)
+                        nm.close_notification(window.last_cmd_end_notification[0])
+                    window.last_cmd_end_notification = None
+                notification_id = nm.notify_with_command(cmd, window.id)
                 if notification_id is not None:
-                    self.last_cmd_end_notification = notification_id, cmd.only_when
+                    window.last_cmd_end_notification = notification_id, cmd.only_when
+
+            if action == 'notify':
+                notify(self, opts, nm)
             elif action == 'bell':
                 self.screen.bell()
+            elif action == 'notify-bell':
+                notify(self, opts, nm)
+                self.screen.bell()
             elif action == 'command':
                 open_cmd([x.replace('%c', self.last_cmd_cmdline).replace('%s', exit_status) for x in notify_cmdline])
             else:
@@ -1758,17 +1797,22 @@ class Window:
     def child_is_remote(self) -> bool:
         for p in self.child.foreground_processes:
             q = list(p['cmdline'] or ())
-            if q and q[0].lower() == 'ssh':
+            if q and os.path.basename(q[0]).lower() == 'ssh':
                 return True
         return False
 
-    def ssh_kitten_cmdline(self) -> list[str]:
+    def ssh_kitten_cmdline_with_pid(self) -> tuple[int, list[str]]:
         from kittens.ssh.utils import is_kitten_cmdline
         for p in self.child.foreground_processes:
             q = list(p['cmdline'] or ())
+            if len(q) > 3 and os.path.basename(q[0]) == 'kitten' and q[1] == 'run-shell':
+                q = q[2:]  # --hold-after-ssh causes kitten run-shell wrapper to be added
             if is_kitten_cmdline(q):
-                return q
-        return []
+                return p['pid'], q
+        return -1, []
+
+    def ssh_kitten_cmdline(self) -> list[str]:
+        return self.ssh_kitten_cmdline_with_pid()[1]
 
     def pipe_data(self, text: str, has_wrap_markers: bool = False) -> PipeData:
         text = text or ''
@@ -1876,6 +1920,7 @@ class Window:
         self.screen.cursor.x = self.screen.cursor.y = 0
         if reset:
             self.screen.reset()
+            self.child.reset_termios_state()
         else:
             self.screen.erase_in_display(3 if scrollback else 2, False)
 
@@ -1883,6 +1928,156 @@ class Window:
         ' Return the last position at which a mouse event was received by this window '
         return get_mouse_data_for_window(self.os_window_id, self.tab_id, self.id)
 
+    # Serialization {{{
+    def as_dict(self, is_focused: bool = False, is_self: bool = False, is_active: bool = False) -> WindowDict:
+        return {
+            'id': self.id,
+            'is_focused': is_focused,
+            'is_active': is_active,
+            'title': self.title,
+            'pid': self.child.pid,
+            'cwd': self.child.current_cwd or self.child.cwd,
+            'cmdline': self.child.cmdline,
+            'last_reported_cmdline': self.last_cmd_cmdline,
+            'last_cmd_exit_status': self.last_cmd_exit_status,
+            'env': self.child.environ or self.child.final_env,
+            'foreground_processes': self.child.foreground_processes,
+            'is_self': is_self,
+            'at_prompt': self.at_prompt,
+            'lines': self.screen.lines,
+            'columns': self.screen.columns,
+            'user_vars': self.user_vars,
+            'created_at': self.created_at,
+            'in_alternate_screen': self.screen.is_using_alternate_linebuf(),
+        }
+
+    def serialize_state(self) -> dict[str, Any]:
+        ans = {
+            'version': 1,
+            'id': self.id,
+            'child_title': self.child_title,
+            'override_title': self.override_title,
+            'default_title': self.default_title,
+            'title_stack': list(self.title_stack),
+            'allow_remote_control': self.allow_remote_control,
+            'remote_control_passwords': self.remote_control_passwords,
+            'cwd': self.child.current_cwd or self.child.cwd,
+            'env': self.child.environ,
+            'cmdline': self.child.cmdline,
+            'last_reported_cmdline': self.last_cmd_cmdline,
+            'last_cmd_exit_status': self.last_cmd_exit_status,
+            'margin': self.margin.serialize(),
+            'user_vars': self.user_vars,
+            'padding': self.padding.serialize(),
+        }
+        if self.window_custom_type:
+            ans['window_custom_type'] = self.window_custom_type
+        if self.overlay_type is not OverlayType.transient:
+            ans['overlay_type'] = self.overlay_type.value
+        if self.user_vars:
+            ans['user_vars'] = self.user_vars
+        return ans
+
+    @property
+    def cwd_for_serialization(self) -> str:
+        cwd = self.get_cwd_of_child(oldest=False) or self.get_cwd_of_child(oldest=True) or self.child.cwd
+        if self.screen.last_reported_cwd and self.at_prompt and not self.child_is_remote:
+            cwd = path_from_osc7_url(self.screen.last_reported_cwd) or cwd
+        return cwd
+
+    def as_launch_command(self, ser_opts: SaveAsSessionOptions, cwd: str, is_overlay: bool = False) -> list[str]:
+        ' Return a launch command that can be used to serialize this window. Empty list indicates not serializable. '
+        if self.actions_on_close or self.actions_on_focus_change or self.actions_on_removal:
+            # such windows are typically UI kittens. The actions are not
+            # serializable anyway, so skip.
+            return []
+        ans = ['launch']
+        if cwd:
+            ans.append(f'--cwd={cwd}')
+        if self.allow_remote_control:
+            ans.append('--allow-remote-control')
+        if self.remote_control_passwords:
+            import shlex
+            for pw, rcp_items in self.remote_control_passwords.items():
+                ans.append(f'--remote-control-password={shlex.join((pw,) + tuple(rcp_items))}')
+        if self.creation_spec:
+            if self.creation_spec.env:
+                for k, v in self.creation_spec.env:
+                    if k not in ('KITTY_PIPE_DATA',):
+                        ans.append(f'--env={k}={v}')
+            for cs in self.creation_spec.colors:
+                ans.append(f'--color={cs}')
+            for wr in self.creation_spec.watchers:
+                ans.append(f'--watcher={wr}')
+            if self.creation_spec.hold:
+                ans.append('--hold')
+            if self.creation_spec.hold_after_ssh:
+                ans.append('--hold-after-ssh')
+        ans.extend(f'--var={k}={v}' for k, v in self.user_vars.items())
+        ans.extend(self.padding.as_launch_args())
+        ans.extend(self.margin.as_launch_args('margin'))
+        if self.override_title:
+            ans.append(f'--title={self.override_title}')
+        wl = get_window_logo_settings_if_not_default(self.os_window_id, self.tab_id, self.id)
+        if wl is not None:
+            logo_path, logo_alpha, logo_pos = wl
+            ans.extend((f'--logo={logo_path}', f'--logo-alpha={logo_alpha}'))
+            xpos = ypos = ''
+            if logo_pos[0] == logo_pos[2] != 0.5:
+                xpos = 'right' if logo_pos[0] else 'left'
+            if logo_pos[1] == logo_pos[3] != 0.5:
+                ypos = 'bottom' if logo_pos[1] else 'top'
+            lpos = 'center'
+            if xpos or ypos:
+                lpos = (f'{ypos}-{xpos}' if ypos else xpos) if xpos else ypos
+            ans.append(f'--logo-position={lpos}')
+
+        if is_overlay:
+            t = 'overlay-main' if self.overlay_type is OverlayType.main else 'overlay'
+            ans.append(f'--type={t}')
+
+        from kittens.ssh.utils import is_kitten_cmdline as is_ssh_kitten_cmdline
+        from kittens.ssh.utils import remove_env_var_from_cmdline, set_cwd_in_cmdline, set_single_env_var_in_cmdline
+        cmd: list[str] = []
+        if self.creation_spec and self.creation_spec.cmd:
+            if self.creation_spec.cmd != resolved_shell(get_options()):
+                cmd = self.creation_spec.cmd
+                if is_ssh_kitten_cmdline(cmd):
+                    if self.at_prompt:
+                        if self.screen.last_reported_cwd:
+                            set_cwd_in_cmdline(path_from_osc7_url(self.screen.last_reported_cwd), cmd)
+        unserialize_data: dict[str, int | list[str] | str] = {'id': self.id}
+        if not cmd and ser_opts.use_foreground_process:
+            def make_exe_absolute(cmd: list[str], pid: int) -> None:
+                if cmd and not os.path.isabs(cmd[0]):
+                    with suppress(Exception):
+                        from .child import abspath_of_exe
+                        cmd[0] = abspath_of_exe(pid)
+            kssh_cmdline = self.ssh_kitten_cmdline()
+            if kssh_cmdline:
+                remove_env_var_from_cmdline('KITTY_SI_RUN_COMMAND_AT_STARTUP', kssh_cmdline)
+                if self.at_prompt:
+                    if self.screen.last_reported_cwd:
+                        set_cwd_in_cmdline(path_from_osc7_url(self.screen.last_reported_cwd), kssh_cmdline)
+                else:
+                    if self.last_cmd_cmdline:
+                        set_single_env_var_in_cmdline('KITTY_SI_RUN_COMMAND_AT_STARTUP', self.last_cmd_cmdline, kssh_cmdline)
+                unserialize_data['cmd_at_shell_startup'] = kssh_cmdline
+            elif not self.at_prompt:
+                if self.last_cmd_cmdline:
+                    unserialize_data['cmd_at_shell_startup'] = self.last_cmd_cmdline
+                elif self.child.pid != (pid := self.child.pid_for_cwd) and pid is not None:
+                    # we have a shell running some command
+                    with suppress(Exception):
+                        fcmd = self.child.cmdline_of_pid(pid)
+                        if fcmd:
+                            make_exe_absolute(fcmd, pid)
+                            unserialize_data['cmd_at_shell_startup'] = fcmd
+        ans.insert(1, unserialize_launch_flag + json.dumps(unserialize_data))
+        ans.extend(cmd)
+        return ans
+    # }}}
+
     # actions {{{
 
     @ac('cp', 'Show scrollback in a pager like less')
@@ -1963,6 +2158,15 @@ class Window:
             self.scroll_end()
             self.write_to_child(self.encoded_key(KeyEvent(key=ord('c'), mods=GLFW_MOD_CONTROL)))
 
+    @ac('cp', 'Copy the selected text from the active window to the clipboard, if no selection,'
+        ' pass the key through to the application running in the terminal.')
+    def copy_or_noop(self) -> bool:
+        text = self.text_for_selection()
+        if text:
+            set_clipboard_string(text)
+            return False
+        return True
+
     @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send SIGINT (aka :kbd:`ctrl+c`)')
     def copy_and_clear_or_interrupt(self) -> None:
         self.copy_or_interrupt()
@@ -2027,18 +2231,25 @@ class Window:
     @ac('sc', '''
         Scroll to the previous/next shell command prompt
         Allows easy jumping from one command to the next. Requires working
-        :ref:`shell_integration`. Takes a single, optional, number as argument which is
-        the number of prompts to jump, negative values jump up and positive values jump down.
-        A value of zero will jump to the last prompt visited by this action.
+        :ref:`shell_integration`. Takes two optional numbers as arguments:
+
+        The first is the number of prompts to jump; negative values jump up and
+        positive values jump down. A value of zero will jump to the last prompt
+        visited by this action. Defaults to -1
+
+        The second is the number of lines to show above the prompt that was
+        jumped to. This is somewhat like `less`'s `--jump-target` option or
+        vim's `scrolloff` setting. Defaults to 0.
+
         For example::
 
-            map ctrl+p scroll_to_prompt -1  # jump to previous
-            map ctrl+n scroll_to_prompt 1   # jump to next
-            map ctrl+o scroll_to_prompt 0   # jump to last visited
+            map ctrl+p scroll_to_prompt -1 3  # jump to previous, showing 3 lines prior
+            map ctrl+n scroll_to_prompt 1     # jump to next
+            map ctrl+o scroll_to_prompt 0     # jump to last visited
         ''')
-    def scroll_to_prompt(self, num_of_prompts: int = -1) -> bool | None:
+    def scroll_to_prompt(self, num_of_prompts: int = -1, scroll_offset: int = 0) -> bool | None:
         if self.screen.is_main_linebuf():
-            self.screen.scroll_to_prompt(num_of_prompts)
+            self.screen.scroll_to_prompt(num_of_prompts, scroll_offset)
             return None
         return True
 
